#! /opt/execute/microperl
################################################################################
# Daylight - Version 1.4 - joghurt(at)southparkkillers.de
# Automatically switches Day/Night-colors, depending on sunrise and -down
#
# Installation: Extract package onto your SD card, root (aka topmost) directory
# If you want to change the switching angle, you have to add
#  "MENUITEM|TASK_SDK1|" to /SdkRegistry/TomTom.mnu if you have a custom menu
#
# I am using TTG v5.420, but maybe this could work with older versions, tell me!
#
# WARNING: Make sure, this scripts uses UNIX newline format! (aka "just LF")
# You better don't edit here if you don't know what this means, just use it...;)
#
# Thanks to:
#  http://lexikon.astronomie.info/zeitgleichung/ for the base formula
################################################################################
# Main
exit if $ARGV[0] eq "SUSPEND"; # No need to save anything...

#use strict;
our $PI = 3.1415926536;
our $CONFIGFILE="daylight/daylight.cfg";
our $TMPFILE="daylight/daylight.tmp";

my $name=$_[split(/\//,$0)-1];$|=1;
my $DEBUG = -z "/mnt/sdcard/debug" || $ARGV[1] =~ /^-*d(ebug)?$/i;

# Load config (angle of dusk and dawn)
my $bak=$/; undef $/;
open(a,"/opt/execute/$CONFIGFILE") || warntheuser("ARGH: $!");
my $config = <a>;
close(a);
$/=$bak;
my $DAWNANGLE=$1 if $config =~ /[\n\r]\s*DAWN\s+([-\.\d]+)/i;
my $DUSKANGLE=$1 if $config =~ /[\n\r]\s*DUSK\s+([-\.\d]+)/i;
my $TELLMEMORE=$1 if $config =~ /[\n\r]\s*TELLMEMORE\s+([-\.\d]+)/i;
$TELLMEMORE |= $DEBUG;

# Possible states here:
# SUSPEND - Do nothing, just kill all daylights...
# BOOT - On Boottime - Calculate angle, switch color and then go sleeping.
# RESUME - When coming up again, needs to do the same as on BOOT
# CONFIGURE - Someone pressed the Button! Calculate actual angle and save it!
#  Then do same as on BOOT or RESUME...
debug("State is: $ARGV[0]");

# First, kill all running daylights... Can't use a pipe, microperl doesn't support that! Sig-Handler also! :(
open(a,"ps -ef |") || warntheuser("ARGH: $!");
my @procs;
foreach my $proc (grep {!/^\s*$$\s/} grep {/$name/} <a>) {
  $proc =~ s/^\s*(\d+).*[\r\n]*$/$1/;
  push(@procs,$proc); # "Prepare to die"...;)
}
close(a);
if ($#procs>-1) {
  debug("Killing: ".join(" ",@procs));
  system("kill ".join(" ",@procs)) ;
}

my $timezoneoffset;
if ($ARGV[0] eq "CONFIGURE") {
  my ($dh,$dm,$ds,$lat,$lon,$dayofyear) = precalc(); # Needed before calculate, loops if no satellite
  switchtoday($dh<12); # If morning switch to day
  # Calculate actual angle and save it!
  my $angle = 30; # Startangle
  my $dist = 512; # First step...
  do {
    my ($dawn,$dusk,$now) = calculate( $dh,$dm,$ds,$lat,$lon,$dayofyear, $angle );
    if ($dh<12 && $now>$dawn || $dh>=12 && $now<$dusk) {
      $angle -= $dist;
    } else {
      $angle += $dist;
    }
    $dist/=2;
  } while ($dist>=1);
  warntheuser("Actual angle: $angle");
  if ($dh<12) {
    $DAWNANGLE = $angle;
    $config =~ s/([\n\r]\s*DAWN\s+)([-\.\d]+)/$1$DAWNANGLE/i;
  } else {
    $DUSKANGLE = $angle;
    $config =~ s/([\n\r]\s*DUSK\s+)([-\.\d]+)/$1$DUSKANGLE/i
  }
  open(a,">/mnt/sdcard/execute/$CONFIGFILE") || warntheuser("ARGH: $!");
  print a $config;
  close(a);
  open(a,">/opt/execute/$CONFIGFILE") || warntheuser("ARGH: $!");
  print a $config;
  close(a);
  sleep 4; # Let the angle be displayed
} else {
  warntheuser("Daylight up and running!") if $TELLMEMORE;
  
  # Preliminary switching on RESUME using TTN-Time
  open(a,"/mnt/flash/settings.dat") || warntheuser("ARGH: $!");
  binmode(a); my $buff; read(a, $buff, 23); read(a, $buff, 2); # Byte 23 and 24
  close(a);
  my $ttnoffset = unpack("v",$buff); # Short, little-endian
  my @ttntime = localtime(time+$ttnoffset*60); # Use system time + TTN-offset (minutes!)
  if (open(a,"/opt/execute/$TMPFILE")) {
    my ($tmplat,$tmplon,$tmpdayofyear,$tmptimezoneoffset) = split(/,/,<a>);
    close(a);
    debug("Found temporary data: $tmplat, $tmplon, $tmpdayofyear, $tmptimezoneoffset");
    my @stime = localtime(time); # Debugging only...
    debug("TTG-time is: ".sprintf("%02d:%02d:%02d",$stime[2],$stime[1],$stime[0]).", TTN-offset is: ".
      sprintf("%02d:%02d",$ttnoffset/60,$ttnoffset%60).", GMT-timezone is: ".
      ($tmptimezoneoffset<0?"-":"+").sprintf("%02d",abs($tmptimezoneoffset)).":00 -> GMT-time is: ".
      sprintf("%02d:%02d:%02d",($ttntime[2]-$tmptimezoneoffset+24)%24,$ttntime[1],$ttntime[0]));
    my ($dawn,$dusk,$now) = calculate( ($ttntime[2]-$tmptimezoneoffset+24)%24,$ttntime[1],$ttntime[0],$tmplat,$tmplon,$tmpdayofyear, (($ttntime[2]-$tmptimezoneoffset+24)%24)<12?$DAWNANGLE:$DUSKANGLE ); # fills calcres
    switchtoday($dawn<$now && $now<$dusk);
  }

  # Switch using Satellites
  precalc() foreach 1..6; # To clear the queue, min. 3 for One, min. 5 for 700!
  my ($dh,$dm,$ds,$lat,$lon,$dayofyear) = precalc(); # Needed before calculate, loops if no satellite
  $timezoneoffset = $ttntime[2]-$dh; # We know ttntime here, not in main loop
  # Using "dawn" before 12 o'clock is not correct, we schould use "dawn" only
  # before sunrise,but we don't know when that is before the calculation...;)
  my ($dawn,$dusk,$now) = calculate( $dh,$dm,$ds,$lat,$lon,$dayofyear, $dh<12?$DAWNANGLE:$DUSKANGLE ); # fills calcres
  switchtoday($dawn<$now && $now<$dusk);
}

# Loop
while (1) {
  my $reloop;
  my $switchto;
  do {
    my ($dh,$dm,$ds,$lat,$lon,$dayofyear) = precalc(); # Needed before calculate, loops if no satellite

    open(a,">/opt/execute/$TMPFILE") || warntheuser("ARGH: $!");
    print a "$lat,$lon,$dayofyear,$timezoneoffset\n";
    close(a);

    my ($dawn,$dusk,$now) = calculate( $dh,$dm,$ds,$lat,$lon,$dayofyear, $dh<12?$DAWNANGLE:$DUSKANGLE ); # fills calcres
    my $sleepdiff;
    if ($now<$dawn) {
      $switchto = 1;
      $sleepdiff = $dawn-$now;
    } elsif ($now<$dusk) {
      $switchto = 0;
      $sleepdiff = $dusk-$now;
    } else {
      $switchto = 1;
      $sleepdiff = 86400+$dawn-$now;
    }
    warntheuser(sprintf("Switching in %02d:%02d:%02d",int($sleepdiff/3600),int($sleepdiff%3600/60),int($sleepdiff%60))) if $DEBUG;
    if ($DEBUG) {
      my $sleeptime = $sleepdiff+$dh*3600+$dm*60+$ds;
      debug(sprintf("It is now: %02d:%02d:%02d, calculated sleeptime is: %02d:%02d:%02d -> %02d:%02d:%02d",
        ($dh+$timezoneoffset)%24,$dm,$ds,int($sleepdiff/3600),int($sleepdiff%3600/60),int($sleepdiff%60),
        (int($sleeptime/3600)+$timezoneoffset)%24,int($sleeptime%3600/60),int($sleeptime%60)));
    }
    # Sleep half the time until 15 minutes before the time
    # For users traveling long distances
    if ($sleepdiff>900) {
      $sleepdiff=int($sleepdiff/2);
      $reloop=1;
    } else {
      $reloop=0;
    }
    debug("Sleeping $sleepdiff...");
    sleep $sleepdiff;
  } while ($reloop);
  warntheuser("Switching Day/Night");
  sleep 6;
  switchtoday($switchto);
}

################################################################################
# Functions
sub precalc() {
  my $gprmc;
  while (1) {
    open(a,"/dev/gpsdata") || warntheuser("ARGH: $!");
    do {
      $gprmc=<a>;
    } while ($gprmc !~ /^\$GPRMC,/);
    close(a);
    debug("$gprmc");
    last if $gprmc =~ /^\$GPRMC,[\d\.]+,A,/;
    sleep 10;
  }
  my @gprmc = (split(/,/,$gprmc))[1,3,4,5,6,9];
  my ($dh,$dm,$ds) = $gprmc[0] =~ /(..)(..)(..)/;
  $dh =~ s/^0//;
  $dm =~ s/^0//;
  $ds =~ s/^0//;
  my $lat = $gprmc[1];
  $lat =~ s!(\d{2})([\d\.]+)!$1+$2/60!e;
  $lat = -$lat if $gprmc[2] eq "S";
  my $lon = $gprmc[3];
  $lon =~ s!(\d{3})([\d\.]+)!$1+$2/60!e;
  $lon = -$lon if $gprmc[4] eq "W";
  my ($ed,$em) = $gprmc[5] =~ /(..)(..)/;
  $ed =~ s/^0//;
  $em =~ s/^0//;
  my @daysofmonths = (0,0,31,59,90,120,151,181,212,243,273,304,334); # Dummy, Jan(0), Feb(31),...
  my $dayofyear=$daysofmonths[$em]+$ed;
  debug("Day of year: $dayofyear - Latitude: $lat - Longitude: $lon");
  return ($dh,$dm,$ds,$lat,$lon,$dayofyear);
}

sub calculate($) {
  my ($dh,$dm,$ds,$lat,$lon,$dayofyear, $sunheight) = @_;
  debug("Using sunheight: $sunheight, GMT-time is ".sprintf("%02d:%02d:%02d",$dh,$dm,$ds));
  my $zone = 0;
  my $rad = $PI/180.0;
  my $h = -($sunheight/60)*$rad;
  my $deklination = 0.40954*sin(0.0172*($dayofyear-79.35));
  my $noon = 12.0*acos((sin($h)-sin($lat*$rad)*sin($deklination))/(cos($lat*$rad)*cos($deklination)))/
    $PI+0.1752*sin(0.033430*$dayofyear+0.5474)+0.1340*sin(0.018234*$dayofyear-0.1939);
  my $u=int((12-$noon-$lon/15.0+$zone)*3600+.5);
  my $d=int((12+$noon-$lon/15.0+$zone)*3600+.5);
  my $n=$dh*3600+$dm*60+$ds;
  return ($u,$d,$n);
}
sub acos($) {
  my $retval;
  if ($_[0] < 0) {
    $retval=$PI-acos(-$_[0]);
  } elsif ($_[0]==0) {
    $retval=$PI/2;
  } else {
    $retval=atan2(sqrt(1-$_[0]*$_[0]),$_[0]);
  }
  return $retval;
}

sub warntheuser($) {
  debug("## FlashMessage: $_[0]");
  display("FlashMessageV01|".$_[0]."|6000",1);
}

sub switchtoday($) {
  # 3=Day color, 2=Night color
  debug("Switching to ".($_[0]?"day":"night"));
  display("ChangeSettingV01|".($_[0]?3:2),2);
}

sub display($$) {
  open(a,">/var/run/SDK.TomTomNavigationServer.$$.$_[1].message"); print a "$_[0]|\0"; close(a);
  open(a,">/var/run/SDK.TomTomNavigationServer.$$.$_[1].finished"); print a "finish\n"; close(a);
}

sub debug($) {
  my $out = $_[0];
  $out =~ s/^[\r\n]*(.*?)[\r\n]*$/$1/;
  print "$$> $out <$$\n" if $DEBUG;
}

################################################################################
